/*******************************************************************************
*
* Fichier de structures pour Mustang .
*
* v.0		02/03/92 .
*******************************************************************************/
typedef struct E_groupe 
{ 
 int label;
 struct E_groupe *next;
} *p_groupe;



extern p_groupe delFromgroup();
extern p_groupe stateGeneration();
extern p_groupe addTogroup();
extern int *choixGroupe();
extern p_groupe mustCodage();
extern void mustPoids();
extern int *PoidsFutur();
extern int *mustCalcul();
extern int *asp_poidEtat();
extern void displayWhiteOut() ;
extern void displayWhiteState() ;
extern void jediCodage();
